# Copyright - Kiven, 2018
from pos_utils import *
import nxt.locator as locator
from nxt.motor import *
import atexit
from nxt.sensor import *
from time import sleep
from threading import Thread
from color_utils import *
from dist_utils import *
import tkinter as tk
from tkinter import Tk


# tunable parameters
GREEN_BLACK_BOUNDARY = 100  # deprecated
GREEN_WHITE_BOUNDARY = 250
_debug = True
TURN_RATIO = 1
LIGHT_BASE = 12
CM_EACH_ROLL = 0
_illuminate = True
RADAR_BASE_INIT = 0.15
RADAR_BASE_PERIODIC = 0.3
RADAR_BASE_POWER = 100

_degree_to_spin_r = to_rolls = \
{
    '90': 0.51,
    '45': 0.26,
    '30': 0.11,
    '15': 0.085,
    '-90': -0.52,
    '-45': -0.265,
    '-30': -0.173,
    '-15': -0.086,
}

def gw(val):  # deprecated
    global GREEN_WHITE_BOUNDARY
    GREEN_WHITE_BOUNDARY = val


def gb(val):  # deprecated
    global GREEN_BLACK_BOUNDARY
    GREEN_BLACK_BOUNDARY = val

print("PyLego initializing.")
print("Copyright - Kiven, 2018")

# initializing fields.
# P.S. *private* fields and functions begin with '_',
# they are not supposed to be invoked by any users (for details please google or baidu it)
# and the author will not be responsible for any mistakes caused by this

brick = None
L = None
R = None
M = None
Radar = None
_lock = False
_stop = False
light = None
sonic = None
touch = None
guard_process = True
color = None
dist = None
pos = Position()  # which marks the position of the robot
boxes = Boxes()  # which stores all the boxes here


''' this file is to be run separately
'''

from tkinter import Tk
import tkinter as tk


def _guard():
    window = Tk()
    def exiter():
        global _stop
        _stop = True
        stop()
        sleep(2)
        import sys
        sys.exit()

    def stopper():
        global _stop

        def work():
            _stop = True
            stop()
            sleep(5)
            _stop = False

        Thread(target=work).start()

    button1 = tk.Button(window, text='stop', command=stopper)
    button2 = tk.Button(window, text='terminate', command=exiter)
    button1.pack()
    button2.pack()
    window.mainloop()


def get_guard():
    return Thread(target=_guard)


def to_cm(r):
    # this converts the roll param to centimeters
    if r > 15:  # todo: complete this
        r /= 360
    return r * CM_EACH_ROLL




'''
def _handle_threads():
    def _do():
        global kill
        kill = True
        sleep(0.1)
        kill = False
    Thread(target=_do()).run()
'''


def l(r=1, p=75, t=None, b=True):  # changed the rule
    sleep(0.2)
    if r < 15:
        r *= 360
    L.turn(p, r, b)


def r(r=1, p=75, t=None, b=True):
    sleep(0.2)
    if r < 15:
        r *= 360
    R.turn(p, r, b)


right = r


def _l(r=1.0, p=100, t=None, b=True):  # changed the rule
    if r < 15:  # todo: say this in the documentation
        r *= 360
    L.turn(p, r, b)


def _r(p=100, r=1, t=None, b=True):
    if r < 15:  # todo: say this in the documentation
        r *= 360
    R.turn(p, r, b)


def _locked(func):
    def output(*args):
        global _lock
        # print("locked")
        _lock = True
        try:
            func(*args)
        except Exception as e:
            print e.message
        finally:
            _lock = False
        # print("unlocked")
    return output


def spin(r=1.0, p=55):
    # L.reset_position(True)
    # R.reset_position(True)
    stop()
    # M.idle()
    sleep(0.1)
    global _lock
    if type(r) == str:
        pos.track(int(r), 0)
        r = to_rolls[r]
    if r < 0:
        r, p = -r, -p
    if r < 15:
        r = int(360*r)  # considered that the _to_roll returns float
    op1 = Thread(target=_locked(L.turn), args=(-p, r, False))
    op2 = Thread(target=_locked(R.turn), args=(p, r, False))
    op1.start()
    op2.start()
    while _lock:
        pass
    stop()
    return L.get_tacho().tacho_count, R.get_tacho().tacho_count
    # print("exit turn")


def hold_on():
    Thread(target=_locked(raw_input),
           args=("To end this hold-on, please enter anything: ",))\
        .start()
    global _lock
    while _lock:
        L.turn(1, 1)
        sleep(0.5)
        R.turn(1, 1)


def _get_counts(motor=None):
    if motor is None:
        motor = M
        return motor.get_tacho().leader_tacho.tacho_count, motor.get_tacho().follower_tacho.tacho_count
    return motor.get_tacho().tacho_count


_temp = 0


def stopped(motor=None):
    if motor is None:
        motor = M
    return motor._get_new_state().power == 0


def _continuous_track():
    global M
    M.reset_position(True)
    prev, _ = _get_counts()
    while not stopped():
        sleep(0.1)
        # print("tracking")
        now, _ = 0, 0
        pos.track(0, to_cm(now-prev))
        prev = now


def f(r=1, p=75, t=None):
    global _lock
    # pos.track(0, to_cm(r))
    if not r or r==0:  # unlimited
        M.run(p)
    else:
        M.turn(p, r if r >= 15 else r*360, False)
        M.brake()
    # return M.get_tacho().leader_tacho.tacho_count, M.get_tacho().follower_tacho.tacho_count


def b(r=1, p=75, t=None):
    f(r, -p, t)


def _test():
    m = SynchronizedMotors(L, R, 0.5)
    m.turn(100, 360)


def stop():
    L.brake()
    R.brake()
    M.brake()
    sleep(0.2)
    L.idle()
    R.idle()
    M.idle()


def _discover(boxes, pos, sonic_dist):
    # this method records a new block
    # if this one is an existed one, return false
    _DELTA_ = 0
    if sonic_dist > 200:
        return False
    robo_pos = pos
    x = robo_pos.x
    y = robo_pos.y
    d = robo_pos.d
    dx = int(sin(rad(d)) * float(sonic_dist))
    dy = int(cos(rad(d)) * float(sonic_dist))
    t_x = x + dx
    t_y = y + dy
    pos = Position(t_x, t_y)
    if boxes.overlapped(pos):
        return False
    else:
        boxes.add(pos)
        return True


def discover(boxes):
    # automatically rotate 360 and record every boxes detected within 200 cm
    global pos
    for i in range(12):
        _discover(boxes, pos, distance())
        spin('30')


def brightness():  # for Task-A this is deprecated, use color() instead
    global LIGHT_BASE
    return light.get_lightness() - LIGHT_BASE


def distance():
    return dist.now


def green():  # this one needs verification
    if color('green'):
        sleep(0.2)
    return color('green')


def black():
    return color('black')


def white():
    return color('white')


def hit():
    return touch.is_pressed() or distance() < 5


def calibrate_light_by_black():  # deprecated
    global LIGHT_BASE
    LIGHT_BASE = light.get_lightness() - 10


def sound():
    Thread(target=brick.play_tone_and_wait, args=(1000, 500)).start()


def _handle_radar_base(motor):
    motor.turn(RADAR_BASE_POWER, RADAR_BASE_INIT)

    def _run(motor):
        while True and not _stop:
            motor.turn(-RADAR_BASE_POWER, RADAR_BASE_PERIODIC)
            motor.turn(RADAR_BASE_POWER, RADAR_BASE_PERIODIC)

    Thread(target=_run, args=(motor,)).start()


def reset(remote=False):
    global brick, L, R, M, radar_base, \
            light, sonic, touch, \
            LIGHT_BASE, _lock,\
            color, dist

    print("CORE: Connecting")
    connect_method = locator.Method(not remote, remote)
    brick = locator.find_one_brick(method=connect_method, debug=True)
    print("Connection to brick established\n")

    print("CORE: Initializing componentss")
    L = Motor(brick, PORT_B)
    R = Motor(brick, PORT_C)
    M = SynchronizedMotors(L, R, TURN_RATIO)

    try:
        radar_base = Motor(brick, PORT_A)
        print("CORE: Found radar turning base")
        _handle_radar_base(radar_base)
    except:
        print("CORE: Found no radar turning base connecting PORT A")
        print("Normal Mode activated")

    _lock = False
    light = Light(brick, PORT_3)
    sonic = Ultrasonic(brick, PORT_2)
    touch = Touch(brick, PORT_4)
    color = Color(light)
    global dist
    dist = Distance(sonic)

    # calibrate light sensor
    light.set_illuminated(True)
    if not _illuminate:
        light.set_illuminated(False)
    sleep(0.5)
    # _LIGHT_BASE_ = brightness() - 10
    print("Initialization completed\n")


try:
    reset()
except locator.BrickNotFoundError:
    print("\nCORE: USB connection not found. Switching connection to bluetooth \n")
    reset(True)

atexit.register(stop)




from os import listdir
from os.path import isfile, join
import numpy as np
from PIL import Image
import sys

def flush(*args):
    # this function clears current line of the console and write something,
    # which helps tracking the progress
    CLEAR = " " * 100 + "\r"
    sys.stdout.write(CLEAR)
    for a in args:
        sys.stdout.write(str(a))  # write() only accepts string
    sys.stdout.flush()


def justfilenames(dir):  # search all files in the dir
    onlyfiles = [f for f in listdir(dir) if isfile(join(dir, f))]
    return onlyfiles


supported = True


try:    # to run safe on Pythonista 3, iOS
    from termcolor import colored
except ImportError:
    print("Warning: without termcolor module, there is no support for colored console output.")
    supported = False


def err(text):  # to print something red
    if supported:
        cont = colored(text, "red")
        print(cont)
    else:
        print(text)



def blue(text):  # to print something blue
    if supported:
        cont = colored(text, "blue")
        print(cont)
    else:
        print(text)


def green(text):  # to print something green
    if supported:
        cont = colored(text, "green")
        print(cont)
    else:
        print(text)


def disablesslvarification():
    ssl._create_default_https_context = ssl._create_unverified_context


def rgb_to_grey_rgbdata(img, size=(64, 64)):
    # PLEASE RESHAPE THE RETURN VALUE TO 1, X, X, 3
    img = img.resize(size, Image.ANTIALIAS)
    grey_scale = img.convert('L')
    one_channel = np.asarray(grey_scale).reshape(64, 64, 1)
    imgdat = np.concatenate((one_channel, one_channel, one_channel), axis=2)
    return imgdat


def main():

    import h5py, numpy as np
    # default: 'a', which means read or write if exists; create otherwise
    f = h5py.File('datasets/train_catvnoncat.h5')
    trainsetx = f['train_set_x']
    print(trainsetx)
    a = np.array(trainsetx)
    print()
    f.close()




if __name__=="__main__":
    main()



'''
now this module maintains a color object
by detecting how fast the brightness is changing
instead of using fixed thresholds
'''

from threading import Thread
from time import sleep

_debug_global = False
_debug_color = True
_power_ref = 0

B_TO_G = 70
G_TO_W = 90
B_TO_W = B_TO_G + G_TO_W


def _handle_green_separately(self):
    if self.color == 'green':
        pass


def _monitor(self):
    print("COLOR: initializing, 1 second to wait")
    sleep(1)
    # well ... I'm pretty sure only ONE second is needed actually
    # because if core module is yet to be loaded, the importation will run into error
    from core import L, R

    while True:
        l_power, r_power = L._get_new_state().power, R._get_new_state().power
        _power = max((abs(l_power), abs(r_power)))
        _r = 1  # sensitivity

        if self.color == 'green':
            pass
        if not self.activate:
            continue
        if l_power * r_power <= 0:  # stopped or spinning
            # continue
            _r = 2
            pass
        if l_power < 0 and r_power < 0:  # backward
            # continue
            pass
        if _power < 50:  # not even moving
            continue
        sleep(0.05)

        _handle_green_separately(self)

        self.history.pop()
        self.history.insert(0, self.now)
        self.prev = sum(self.history[4:]) / (len(self.history)-4)
        self.reset()
        self.now = sum(self.history[4:]) / (len(self.history)-4)

        if _debug_color and abs(self.prev - self.now) > 12:
            print "COLOR value change: ", self.prev, ' to ', self.now, ', now', self.color.upper()

        # GETTING BRIGHTER
        if self.now - self.prev > _r*G_TO_W and self.color == 'green':
                self.color = 'white'
                self.reset()
                print "COLOR: change from GREEN 2 WHITE"
        elif _r*B_TO_G < self.now- self.prev < _r*B_TO_W and self.color == 'black':
                self.color = 'green'
                self.reset()
                print "COLOR: change from BLACK 2 GREEN"
        elif _r*B_TO_W < self.now - self.prev and self.color == 'black':
            self.color = 'white'
            self.reset()
            print "COLOR: change from BLACK 2 WHITE"

        # GETTING DARKER
        if _r*B_TO_W > self.prev - self.now > G_TO_W*_r and self.color == 'white':
            self.color = 'green'
            self.reset()
            print("COLOR: change from WHITE 2 GREEN")
        elif _r*B_TO_G < self.prev - self.now < _r*G_TO_W and self.color == 'green':
            self.color = 'black'
            self.reset()
            print("COLOR: change from GREEN 2 BLACK")
        elif _r*B_TO_W < self.prev - self.now:
            if self.color == 'white':
                self.color = 'black'
                self.reset()
                print("COLOR: change from WHITE 2 BLACK")


class Color:
    def __init__(self, sensor, calibrate_by='white'):
        self.sensor = sensor
        self.history = [sensor.get_lightness() for i in range(10)]
        self.now = sensor.get_lightness()
        self.color = calibrate_by
        self.activate = True
        work = Thread(target=_monitor, args=(self,))
        work.start()

    def reset(self):
        def _work(self):
            self.history = [(self.sensor.get_lightness(),
                             sleep(0.01)
                             )[0] for i in range(10)]
        _work(self)
        # Thread(target=_work, args=(self,)).start()

    def __call__(self, *args, **kwargs):
        if args != ():
            return any([i == self.color for i in args])
        return self.color



from core import *
import core
import atexit
atexit.register(stop)


def run():
	# tunable parameters
	kp = 1000
	ki = 100
	kd = 10000
	interval = 0.05
	offset = (core.GREEN_BLACK_BOUNDARY + core.GREEN_WHITE_BOUNDARY) / 2 - 5 # todo: mean of brightness

	# non-tunnable parameters
	tp = 50
	integral = 0
	lasterror = 0

	# main loop
	while True:
		light = brightness()
		error = light - offset
		integral += error
		deriv = error - lasterror
		turn = kp*error + ki*integral + kd*deriv
		turn /= 100
		powerL = tp + turn
		powerR = tp - turn
		print(powerL, ' ', powerR)
		L.run(powerL)
		R.run(powerR)
		lasterror = error
		sleep(interval)

run()